//------------//------------//------------//------------//------------//------------
//
// cl rebeffect.cpp /EHsc /Ox
//
//------------//------------//------------//------------//------------//------------
#include <vector>
#include <stdio.h>
#include <math.h>
#include <windows.h>
#include <process.h>
#include <mmsystem.h>
#include <mmreg.h>

//------------//------------//------------//------------//------------//------------
// comment
//------------//------------//------------//------------//------------//------------
#pragma comment(lib,     "winmm.lib")
#pragma comment(lib,     "user32.lib")
#pragma comment(lib,     "ole32.lib")

//------------//------------//------------//------------//------------//------------
// Def
//------------//------------//------------//------------//------------//------------
typedef float t_snd;
//typedef short t_snd;
typedef void (*SoundFillFunc)(void *dest, void *src, int samples);

//------------//------------//------------//------------//------------//------------
// Config
//------------//------------//------------//------------//------------//------------
enum {
	Bits         = (sizeof(t_snd) * 8),
	Channel      = (2),
	Freq         = (44100),
	Align        = ((Channel * Bits) / 8),
	BytePerSec   = (Freq * Align),
	BufNum       = 4,
	Samples      = 1024,
};


struct SoundCallBack {
	SoundFillFunc inf;
	SoundFillFunc outf;
};

//------------//------------//------------//------------//------------//------------
// SoundThread
//------------//------------//------------//------------//------------//------------
static VOID SoundThreadRecord(void *func) {
	enum {
		SoundIn  = 0,
		SoundOut,
		SoundMax,
	};
	SoundCallBack  *pcb = (SoundCallBack *)func;
	HWAVEIN        hwi               = NULL;
	HWAVEOUT       hwo               = NULL;
	DWORD          countin             = 0;
	DWORD          countout          = 0;
	SoundFillFunc  infill            = pcb->inf;
	SoundFillFunc  outfill           = pcb->outf;
	HANDLE         hEvent[SoundMax]  = {
		CreateEvent(NULL, FALSE, FALSE, NULL),
		CreateEvent(NULL, FALSE, FALSE, NULL),
	};

	WAVEFORMATEX   wfx               = {
		sizeof(t_snd) == sizeof(float) ? WAVE_FORMAT_IEEE_FLOAT : WAVE_FORMAT_PCM,
		Channel, Freq, BytePerSec, Align, Bits, 0
	};
	WAVEHDR        whdrin[BufNum];
	WAVEHDR        whdrout[BufNum];

	waveInOpen (&hwi, WAVE_MAPPER, &wfx, (DWORD)hEvent[SoundIn], 0, CALLBACK_EVENT);
	waveOutOpen(&hwo, WAVE_MAPPER, &wfx, (DWORD)hEvent[SoundOut], 0, CALLBACK_EVENT);
	std::vector< std::vector<char> > soundbuffer;
	std::vector< std::vector<char> > soundbufferout;
	soundbuffer.resize(BufNum);
	soundbufferout.resize(BufNum);
	for(int i = 0 ; i < BufNum; i++) {
		soundbuffer[i].resize(Samples * wfx.nBlockAlign);
		soundbufferout[i].resize(Samples * wfx.nBlockAlign);
		WAVEHDR tempin  = { &soundbuffer[i][0], Samples * wfx.nBlockAlign, 0, 0, 0, 0, NULL, 0 };
		WAVEHDR tempout = { &soundbufferout[i][0], Samples * wfx.nBlockAlign, 0, 0, 0, 0, NULL, 0 };
		whdrin[i]       = tempin;
		whdrout[i]      = tempout;
		waveInPrepareHeader(hwi, &whdrin[i], sizeof(WAVEHDR));
		waveInAddBuffer(hwi, &whdrin[i], sizeof(WAVEHDR));
	}
	
	//Record Start
	waveInStart(hwi);

	
	//Start MSG
	MSG	 msg;
	while(TRUE) {
		if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
			if(msg.message == WM_QUIT) break;
		
		DWORD dwSignal = WaitForMultipleObjects(SoundMax, hEvent, FALSE, 1000);
		
		//IN
		if(dwSignal == WAIT_OBJECT_0 + 0) {
			if(whdrin[countin].dwFlags & WHDR_DONE) {
				infill((void *)whdrout[countin].lpData, whdrin[countin].lpData, whdrin[countin].dwBufferLength);
				waveOutPrepareHeader(hwo, &whdrout[countout], sizeof(WAVEHDR));
				waveOutWrite(hwo, &whdrout[countout], sizeof(WAVEHDR));
				waveInPrepareHeader(hwi, &whdrin[countin], sizeof(WAVEHDR));
				waveInAddBuffer(hwi, &whdrin[countin], sizeof(WAVEHDR));
				countout = (countout + 1) % BufNum;
				countin  = (countin + 1) % BufNum;
			}
		}

		//OUT
		if(dwSignal == WAIT_OBJECT_0 + 1) {
		}
	}

	//in
	do {
		countin = 0;
		for(int i = 0; i < BufNum; i++)
			countin += (whdrin[i].dwFlags & WHDR_DONE) ? 0 : 1;
		if(countin) Sleep(50);
	} while(countin);
	
	//out
	do {
		countout = 0;
		for(int i = 0; i < BufNum; i++)
			countout += (whdrout[i].dwFlags & WHDR_DONE) ? 0 : 1;
		if(countout) Sleep(50);
	} while(countout);
	

	for(int i = 0 ; i < BufNum ; i++) {
		waveInUnprepareHeader(hwi, &whdrin[i], sizeof(WAVEHDR));
		waveOutUnprepareHeader(hwo, &whdrout[i], sizeof(WAVEHDR));
	}

	waveInReset(hwi);
	waveOutReset(hwo);
	waveInClose(hwi);
	waveOutClose(hwo);
	
	for(int i = 0 ; i < SoundMax; i++) {
		if(hEvent) CloseHandle(hEvent);
	}
}

//------------//------------//------------//------------//------------//------------
// SoundInit
//------------//------------//------------//------------//------------//------------
HANDLE SoundInit(SoundFillFunc inFunc, SoundFillFunc outFunc) {
	SoundCallBack cb = {inFunc, outFunc};
	return (HANDLE)_beginthread(SoundThreadRecord, 0, &cb);
}

//------------//------------//------------//------------//------------//------------
// SoundTerm
//------------//------------//------------//------------//------------//------------
void SoundTerm(HANDLE hThread) {
	if(!hThread) return;
	PostThreadMessage(GetThreadId(hThread), WM_QUIT, 0, 0);
	WaitForSingleObject(hThread, 5000); //max5s
	CloseHandle(hThread);
}


//------------//------------//------------//------------//------------//------------
//
// Reverb
//
//------------//------------//------------//------------//------------//------------
struct DelayBuffer {
	enum {
		Max = 100000,
	};
	float Buf[Max];
	unsigned long	 Rate;
	unsigned int	 Index;
	void Init(unsigned long m)				{ Rate = m;	}
	void Update(float a)							{ Buf[Index++ % Rate ] = a;	}
	float Sample(unsigned long n = 0) { return Buf[ (Index + n) % Rate]; }
};

//http://www.ari-web.com/service/soft/reverb-2.htm
struct Reverb {
	enum {
		CombMax = 8,
		AllMax	= 2,
	};
	DelayBuffer comb[CombMax];
	DelayBuffer all[AllMax];

	float Sample(float a, int index = 0, int character = 0, int lpfnum = 4) {
		const int tau[][4][4] = {
			//prime
			{
				{2063, 1847, 1523, 1277}, {3089, 2927, 2801, 2111}, {5479, 5077, 4987, 4057}, {9929, 7411, 4951, 1063},
			},
			{
				{2053, 1867, 1531, 1259}, {3109, 2939, 2803, 2113}, {5477, 5059, 4993, 4051}, {9949, 7393, 4957, 1097},
			}
		};

		const float gain[] = 
		{
			 -(0.8733), -(0.8223), -(0.8513), -(0.8503),
		};
		float D = a * 0.5;
		float E = 0;
		
		//Comb
		for(int i = 0 ; i < CombMax; i++)
		{
			DelayBuffer *reb = &comb[i];
			reb->Init(tau[character % 2][index % 4][i]);
			float k = 0;
			float c = 0;
			int LerpMax = lpfnum + 1;
			for(int h = 0 ; h < LerpMax; h++)
				k += reb->Sample(h * 2);
			k /= float(LerpMax);
			c = a + k;
			reb->Update(c * gain[i] * 1.1);
			E += c;
		}
		D = (D + E) * 0.3;
		return D;
	}
};



//------------//------------//------------//------------//------------//------------
//
// User
//
//------------//------------//------------//------------//------------//------------
Reverb rebL;
Reverb rebR;
DelayBuffer delayL;
DelayBuffer delayR;
static int state = 0;

//fill callback
void inFunc(void *dest, void *src, int samples)
{
  t_snd *s = (t_snd *)src;
  t_snd *d = (t_snd *)dest;
  samples /= sizeof(t_snd);
  samples /= Channel;
  
  //reverb san
  if(state == 0) {
    for(int i = 0; i < samples; i++)
    {
      t_snd v = s[i * 2 + 0] * 0.5;
      d[i * 2 + 0] = 0.5 * v + (rebL.Sample(v, 0, 0, 4));
      d[i * 2 + 1] = 0.5 * v + (rebR.Sample(v, 0, 1, 4));
    }
  }
  
  //delay
  if(state == 1) {
    float L = 0; 
    float R = 0; 
    delayL.Init(12237);
    delayR.Init(12236);
    for(int i = 0; i < samples; i++)
    {
      t_snd v = s[i * 2 + 0];
      d[i * 2 + 0] = delayL.Sample() + v;
      d[i * 2 + 1] = delayR.Sample() + v;
      delayL.Update(d[i * 2 + 0] * 0.5);
      delayR.Update(d[i * 2 + 1] * 0.5);
    }
  }
  
  //hutuu
  if(state == 2) {
    float L = 0; 
    float R = 0; 
    for(int i = 0; i < samples; i++) {
      t_snd v = s[i * 2 + 0];
      d[i * 2 + 0] = v;
      d[i * 2 + 1] = v;
    }
  }
  
  printf("%3.4f %3.4f\r", abs(d[0]), abs(d[1]));
}

void outFunc(void *dest, void *src, int samples)
{
}

//------------//------------//------------//------------//------------//------------
// ep
//------------//------------//------------//------------//------------//------------
int main(int argc, char *argv[])
{
	printf("Key1:Reverb, Key2:Delay, Key3:Normal\n");
	HANDLE h = SoundInit(inFunc, outFunc);
	while(h && !(GetAsyncKeyState(VK_ESCAPE) & 0x8000)) {
		if(GetAsyncKeyState('1') & 0x8000) {
			state = 0;
		}
		if(GetAsyncKeyState('2') & 0x8000) {
			state = 1;
		}
		if(GetAsyncKeyState('3') & 0x8000) {
			state = 2;
		}
		
		
		Sleep(16);
	}
	SoundTerm(h);
	return 0;
}

